import * as b from 'bobril';
import * as css from '../../theme/css/css';
import * as colors from '../../common/colors';
import * as constants from '../../common/constants';

const hintStyle = b.styleDef({
    color: colors.defaultWhite,
    backgroundColor: colors.redHighlight,
    position: 'absolute',
    top: 0,
    left: 0,
    strokeOpacity: 0.4
});

export interface IHintData {
    children?: b.IBobrilChildren;
    top?: number;
    left?: number;
    countedTop?: number;
    countedLeft?: number;
    moveTop?: number;
    moveLeft?: number;
    oldCountedTop?: number;
    oldCountedLeft?: number;
    color?: string;
    backgroundColor?: string;
    waitForAnimation?: number; // in miliseconds
    onMove?: (dx: number, dy: number) => void;
    fullScreen: boolean;
}

interface IHintCtx extends b.IBobrilCtx {
    data: IHintData;
    opac: number;
    lastX: number;
    lastY: number;
    newHeight: number;
    newWidth: number;
    oldHeight: number;
    oldWidth: number;
    oldFullScreen: boolean;
    timeout: number;
    inter: number;
    eventListener: () => void;
}

export const hint = b.createComponent<IHintData>({
    id: 'hint',
    init(ctx: IHintCtx, me: b.IBobrilNode) {
        if (ctx.data.waitForAnimation !== undefined) {
            ctx.opac = 0.005;
        }
    },
    render(ctx: IHintCtx, me: b.IBobrilNode) {
        me.children = ctx.data.children;

        b.style(
            me,
            hintStyle,
            css.well,
            !!ctx.data.top && { top: ctx.data.countedTop },
            !!ctx.data.left && { left: ctx.data.countedLeft },
            !!ctx.data.color && { color: ctx.data.color },
            !!ctx.data.backgroundColor && { backgroundColor: ctx.data.backgroundColor },
            { opacity: ctx.data.waitForAnimation === undefined ? 1 : ctx.opac }
        );

        function createInterval() {
            ctx.inter = setInterval(
                () => {
                    ctx.opac += (0.04 / constants.slowedTime);
                    b.invalidate(ctx);
                    if (Math.floor(ctx.opac) === 1) {
                        document.body.removeEventListener('finish-animation', ctx.eventListener, false);
                        clearTimeout(ctx.timeout);
                        ctx.timeout = 0;
                        clearInterval(ctx.inter);
                        ctx.inter = 0;
                    }
                },
                100);
        }

        if (ctx.opac === 0.005 && !ctx.timeout) {
            ctx.timeout = setTimeout(createInterval, ctx.data.waitForAnimation * constants.slowedTime);

            ctx.eventListener = () => {
                ctx.opac = 1;
                clearTimeout(ctx.timeout);
                ctx.timeout = 0;
                clearInterval(ctx.inter);
                ctx.inter = 0;
                document.body.removeEventListener('finish-animation', ctx.eventListener, false);
                b.invalidate(ctx);
            };

            document.body.addEventListener('finish-animation', ctx.eventListener, false);
        }

        ctx.oldHeight = ctx.newHeight;
        ctx.oldWidth = ctx.newWidth;
    },
    onPointerDown(ctx: IHintCtx, param: b.IBobrilPointerEvent) {
        b.registerMouseOwner(ctx);
        ctx.lastX = param.x;
        ctx.lastY = param.y;
        return true;
    },
    onPointerMove(ctx: IHintCtx, param: b.IBobrilPointerEvent) {
        if (!b.isMouseOwner(ctx)) {
            return false;
        }

        if (ctx.lastX === param.x && ctx.lastY === param.y) {
            return false;
        }

        if (ctx.data.onMove) {
            let svg = document.getElementsByTagName('svg')[0];

            let x = param.x - ctx.lastX;
            let y = param.y - ctx.lastY;

            if (svg !== undefined) {
                x = ((x * constants.width) / svg.getBoundingClientRect().width);
                y = ((y * constants.height) / svg.getBoundingClientRect().height);
            }

            ctx.data.onMove(x, y);
        }

        ctx.lastX = param.x;
        ctx.lastY = param.y;
        return true;
    },
    onPointerUp(ctx: IHintCtx, param: b.IBobrilPointerEvent) {
        b.releaseMouseOwner();
        return true;
    },
    onPointerCancel(ctx: IHintCtx, param: b.IBobrilPointerEvent) {
        return true;
    }
});