import * as b from 'bobril';
import * as colors from '../common/colors';
import * as constants from '../common/constants';
import { IScene } from '../model/iScene';
import { IVertex } from '../model/iVertex';
import { IHint } from '../model/iHint';
import { IItem } from '../model/iItem';
import { IEdge } from '../model/iEdge';
import { IPoint } from '../model/iPoint';
import { IAction } from '../model/iAction';
import { IHalfCircle } from '../model/iHalfCircle';

export function addVertex(
    id: string,
    point: IPoint,
    label: string,
    strokeColor?: string,
    strokeWidth?: number,
    color?: string,
    move?: boolean): IAction {
    return {
        update: scene => scene.vertices.push({ id, point, label, strokeColor, strokeWidth, color, move }),
        rollback: scene => removeById(scene.vertices, id)
    };
}

export function checkProperties<T, K>(source: T, dest: K) {
    Object.keys(source).forEach(key => {
        if (source[key] !== undefined)
            dest[key] = source[key];
    });
}

export function editVertex(
    id: string,
    point?: IPoint,
    label?: string,
    strokeColor?: string,
    strokeWidth?: number,
    color?: string,
    move?: boolean,
    waitForAnimation?: number): IAction {
    let originalVertex: IVertex = null;
    return {
        update: scene => {
            originalVertex = getVertexById(scene, id);
            const index = scene.vertices.indexOf(originalVertex);

            if (waitForAnimation !== undefined) {
                let timeout = setTimeout(
                    () => {
                        edit();
                        clearTimeout(timeout);
                        document.body.removeEventListener('finish-animation', eventListener, false);
                    },
                    waitForAnimation * constants.slowedTime);

                let eventListener = () => {
                    edit();
                    clearTimeout(timeout);
                    document.body.removeEventListener('finish-animation', eventListener, false);
                };

                document.body.addEventListener('finish-animation', eventListener, false);
            } else {
                edit();
            }

            function edit() {
                scene.vertices[index] = b.assign({}, originalVertex);
                checkProperties(
                    {
                        id,
                        point,
                        label,
                        strokeColor,
                        strokeWidth,
                        color,
                        move,
                        waitForAnimation
                    },
                    scene.vertices[index]
                );
                b.invalidate();
            }
        },
        rollback: scene => {
            removeById(scene.vertices, id);
            scene.vertices.push(originalVertex);
        }
    };
}

export function highlightVertex(
    id: string,
    strokeWidth: number = 7,
    strokeColor: string = colors.redHighlight): IAction {
    let originalVertex: IVertex = null;
    return {
        update: scene => {
            originalVertex = removeById(scene.vertices, id);
            scene.vertices.push({
                id, point: originalVertex.point, label: originalVertex.label,
                strokeColor, strokeWidth, color: originalVertex.color, move: originalVertex.move
            });
        },
        rollback: scene => {
            removeById(scene.vertices, id);
            scene.vertices.push(originalVertex);
        }
    };
}

export function unHighlightVertex(id: string): IAction {
    let originalVertex: IVertex = null;
    return {
        update: scene => {
            originalVertex = removeById(scene.vertices, id);
            scene.vertices.push({ id, point: originalVertex.point, label: originalVertex.label });
        },
        rollback: scene => {
            removeById(scene.vertices, id);
            scene.vertices.push(originalVertex);
        }
    };
}

export function addCloud(): IAction {
    return {
        update: scene => scene.cloud = true,
        rollback: scene => scene.cloud = false
    };
}

export function removeCloud(): IAction {
    return {
        update: scene => scene.cloud = false,
        rollback: scene => scene.cloud = true
    };
}

export function addEdge(
    id: string,
    startPoint: IPoint | string,
    endPoint: IPoint | string,
    interpolation?: number,
    strokeWidth?: number,
    color?: string,
    dotted?: boolean,
    priority?: number,
    moved?: number,
    waitForAnimation?: number
): IAction {
    return {
        update: scene => scene.edges.push({
            id,
            startPoint: getPointFromPointOrVertexId(scene, startPoint),
            endPoint: getPointFromPointOrVertexId(scene, endPoint),
            interpolation,
            strokeWidth,
            color,
            dotted,
            priority,
            moved,
            waitForAnimation
        }),
        rollback: scene => removeById(scene.edges, id)
    };
}

export function addImage(
    id: string,
    point: IPoint,
    width: number,
    height: number,
    type: string,
    waitForAnimation?: number): IAction {
    return {
        update: scene => scene.images.push({
            id,
            point,
            width,
            height,
            type,
            waitForAnimation
        }),
        rollback: scene => removeById(scene.images, id)
    };
}

export function addUnspecificGraph(
    id: string,
    startPoint: IPoint,
    endPoint: IPoint,
    priority?: number): IAction {
    const addVertexAction = addVertex(id, startPoint, '');
    const addEdgeAction = addEdge(id + '-line', startPoint, endPoint, undefined, undefined, undefined, true);
    return {
        update: scene => {
            addVertexAction.update(scene);
            addEdgeAction.update(scene);
        },
        rollback: scene => {
            addEdgeAction.rollback(scene);
            addVertexAction.rollback(scene);
        }
    };
}

export enum HighlightType {
    Start = 0,
    End = 1,
    Both = 2,
    Full = 3
}

export function highlightEdge(
    id: string,
    type: HighlightType,
    strokeWidth: number = 7,
    color: string = colors.redHighlight,
    dotted?: boolean,
    moved?: number,
    waitForAnimation?: number,
    newId?: string,
    priority?: number,
): IAction {
    return {
        update: scene => {
            const edge = getEdgeById(scene, id);

            switch (type) {
                case HighlightType.Start:
                    scene.edges.push({
                        id: id + '-highlight-start',
                        startPoint: edge.startPoint,
                        endPoint: edge.endPoint,
                        interpolation: 1 / 3,
                        strokeWidth,
                        color,
                        priority: priority === undefined ? 50 : priority
                    });
                    break;
                case HighlightType.End:
                    scene.edges.push({
                        id: id + '-highlight-end',
                        startPoint: edge.endPoint,
                        endPoint: edge.startPoint,
                        interpolation: 1 / 3,
                        strokeWidth,
                        color,
                        priority: priority === undefined ? 50 : priority
                    });
                    break;
                case HighlightType.Both:
                    scene.edges.push({
                        id: id + '-highlight-start',
                        startPoint: edge.startPoint,
                        endPoint: edge.endPoint,
                        interpolation: 1 / 3,
                        strokeWidth,
                        color,
                        priority: priority === undefined ? 50 : priority
                    });
                    scene.edges.push({
                        id: id + '-highlight-end',
                        startPoint: edge.endPoint,
                        endPoint: edge.startPoint,
                        interpolation: 1 / 3,
                        strokeWidth,
                        color,
                        priority: priority === undefined ? 50 : priority
                    });
                    break;
                case HighlightType.Full:
                    scene.edges.push({
                        id: (newId || id) + '-highlight-full',
                        startPoint: edge.startPoint,
                        endPoint: edge.endPoint,
                        strokeWidth: strokeWidth,
                        color: color,
                        dotted: dotted,
                        priority: priority === undefined ? 50 : priority,
                        moved: moved,
                        waitForAnimation: waitForAnimation,
                    });
                    break;
            }
        },
        rollback: scene => {
            switch (type) {
                case HighlightType.Start:
                    removeById(scene.edges, id + '-highlight-start');
                    break;
                case HighlightType.End:
                    removeById(scene.edges, id + '-highlight-end');
                    break;
                case HighlightType.Both:
                    removeById(scene.edges, id + '-highlight-start');
                    removeById(scene.edges, id + '-highlight-end');
                    break;
                case HighlightType.Full:
                    removeById(scene.edges, (newId || id) + '-highlight-full');
                    break;
            }
        }
    };
}

export function unHighlightEdge(id: string, type: HighlightType): IAction {
    let originalEdges: IEdge[] = null;
    return {
        update: scene => {
            originalEdges = [];
            switch (type) {
                case HighlightType.Start:
                    originalEdges.push(removeById(scene.edges, id + '-highlight-start'));
                    break;
                case HighlightType.End:
                    originalEdges.push(removeById(scene.edges, id + '-highlight-end'));
                    break;
                case HighlightType.Both:
                    originalEdges.push(removeById(scene.edges, id + '-highlight-start'));
                    originalEdges.push(removeById(scene.edges, id + '-highlight-end'));
                    break;
                case HighlightType.Full:
                    originalEdges.push(removeById(scene.edges, id + '-highlight-full'));
                    break;
            }
        },
        rollback: scene => {
            originalEdges.forEach(edge => {
                edge.waitForAnimation = undefined;
                scene.edges.push(edge);
            });
        }
    };
}

export function addHint(
    id: string,
    point: IPoint,
    text: b.IBobrilChildren,
    color?: string,
    backgroundColor?: string,
    waitForAnimation?: number,
    countedPoint?: IPoint,
    oldCountedPoint?: IPoint,
): IAction {
    return {
        update: scene => scene.hints.push({
            id, point: point, text, color, backgroundColor, waitForAnimation,
            countedPoint,
            oldCountedPoint
        }),
        rollback: scene => {
            removeById(scene.hints, id);
        }
    };
}

export function addAxis(
    id: string): IAction {
    return {
        update: scene => scene.axis = { id },
        rollback: scene => scene.axis = undefined
    };
}

export function addHalfCircle(
    id: string,
    point: IPoint,
    rotation?: boolean,
    color?: string,
    strokeColor?: string,
    strokeWidth?: number,
    priority?: number): IAction {
    return {
        update: scene => scene.halfCircles.push({
            id, point: point, rotation: rotation, color: color,
            strokeColor: strokeColor, strokeWidth: strokeWidth, priority: priority
        }),
        rollback: scene => removeById(scene.halfCircles, id)
    };
}

export function editWhiteboard(
    id: string,
    point?: IPoint,
    text?: b.IBobrilChildren,
    color?: string,
    backgroundColor?: string,
    margin?: number,
    waitForAnimation?: number): IAction {
    let originalWhiteboard: IHint = null;
    return {
        update: scene => {
            originalWhiteboard = getWhiteboardById(scene, id);
            const index = scene.whiteboards.indexOf(originalWhiteboard);
            scene.whiteboards[index] = b.assign({}, originalWhiteboard);
            checkProperties(
                {
                    id,
                    point,
                    text,
                    color,
                    backgroundColor,
                    margin,
                    waitForAnimation
                },
                scene.whiteboards[index]
            );
        },
        rollback: scene => {
            removeById(scene.whiteboards, id);
            scene.whiteboards.push(originalWhiteboard);
        }
    };
}

export function addWhiteboard(
    id: string,
    point: IPoint,
    text: b.IBobrilChildren,
    color?: string,
    backgroundColor?: string,
    margin?: number,
    waitForAnimation?: number,
    fontSize?: number,
    countedPoint?: IPoint,
    oldCountedPoint?: IPoint, ): IAction {
    return {
        update: scene => scene.whiteboards.push({
            id, point: point, text, color, backgroundColor,
            margin, waitForAnimation, fontSize, countedPoint, oldCountedPoint
        }),
        rollback: scene => removeById(scene.whiteboards, id)
    };
}

export function addText(
    id: string,
    point: IPoint,
    text: b.IBobrilChildren,
    color?: string,
    backgroundColor?: string,
    size?: number): IAction {
    return {
        update: scene => scene.texts.push({ id, point: point, text, color, backgroundColor, size }),
        rollback: scene => removeById(scene.texts, id)
    };
}

export function removeHint(id: string): IAction {
    let originalHint: IHint = null;
    return {
        update: scene => originalHint = removeById(scene.hints, id),
        rollback: scene => {
            originalHint.waitForAnimation = undefined;
            scene.hints.push(originalHint);
        }
    };
}

export function removeEdge(id: string): IAction {
    let originalEdge: IEdge = null;
    return {
        update: scene => originalEdge = removeById(scene.edges, id),
        rollback: scene => {
            originalEdge.waitForAnimation = undefined;
            scene.edges.push(originalEdge);
        }
    };
}

export function removeVertex(id: string): IAction {
    let originalVertex: IVertex = null;
    return {
        update: scene => originalVertex = removeById(scene.vertices, id),
        rollback: scene => {
            scene.vertices.push(originalVertex);
        }
    };
}

export function removeUnspecificGraph(id: string): IAction {
    let originalEdge: IEdge = null;
    let originalVertex: IVertex = null;
    return {
        update: scene => {
            originalVertex = removeById(scene.vertices, id);
            originalEdge = removeById(scene.edges, id + '-line');
        },
        rollback: scene => {
            scene.vertices.push(originalVertex);
            scene.edges.push(originalEdge);
        }
    };
}

export function editHint(
    id: string,
    point?: IPoint,
    text?: b.IBobrilChildren,
    color?: string,
    backgroundColor?: string): IAction {
    let originalHint: IHint = null;
    return {
        update: scene => {
            originalHint = getHintById(scene, id);
            const index = scene.hints.indexOf(originalHint);
            scene.hints[index] = b.assign({}, originalHint);
            checkProperties(
                {
                    id,
                    point,
                    text,
                    color,
                    backgroundColor,
                },
                scene.hints[index]
            );
        },
        rollback: scene => {
            removeById(scene.hints, id);
            scene.hints.push(originalHint);
        }
    };
}

export function editHalfCircle(
    id: string,
    point?: IPoint,
    rotation?: boolean,
    color?: string,
    strokeColor?: string,
    strokeWidth?: number,
    priority?: number): IAction {
    let originalHalfCircle: IHalfCircle = null;
    return {
        update: scene => {
            originalHalfCircle = getItemById(scene.halfCircles, id);
            const index = scene.halfCircles.indexOf(originalHalfCircle);
            scene.halfCircles[index] = b.assign({}, originalHalfCircle);
            checkProperties(
                {
                    id,
                    point,
                    rotation,
                    color,
                    strokeColor,
                    strokeWidth,
                    priority
                },
                scene.halfCircles[index]
            );
        },
        rollback: scene => {
            removeById(scene.halfCircles, id);
            scene.halfCircles.push(originalHalfCircle);
        }
    };
}

export function removeWhiteboard(id: string): IAction {
    let whiteboard: IHint = null;
    return {
        update: scene => whiteboard = removeById(scene.whiteboards, id),
        rollback: scene => {
            whiteboard.waitForAnimation = undefined;
            scene.whiteboards.push(whiteboard);
        }
    };
}

export function removeItem<TItem extends IItem>(id: string, array: (scene: IScene) => TItem[]): IAction {
    let originalItem: TItem = null;
    return {
        update: scene => originalItem = removeById(array(scene), id),
        rollback: scene => {
            array(scene).push(originalItem);
        }
    };
}

export function removeById<TITem extends IItem>(items: TITem[], id: string): TITem {
    const item = items.filter(i => i.id === id)[0];
    const index = items.indexOf(item);
    if (index >= 0) {
        items.splice(index, 1);
    } else {
        throw `${id} je požadován před jeho přidáním do scény`;
    }

    return item;
}

export function getPointFromPointOrVertexId(scene: IScene, pointOrId: IPoint | string): IPoint {
    if (typeof pointOrId === 'string') {
        const vertex = scene.vertices.filter(v => v.id === pointOrId)[0];
        if (!vertex)
            throw `Vertex ${pointOrId} je požadován před jeho přidáním do scény`;

        return vertex.point;
    } else {
        return pointOrId;
    }
}

export function getEdgeById(scene: IScene, id: string): IEdge {
    const edge = scene.edges.filter(e => e.id === id)[0];
    if (!edge)
        throw `Edge ${id} je požadován před jeho přidáním do scény`;

    return edge;
}

export function getHintById(scene: IScene, id: string): IHint {
    const hint = scene.hints.filter(e => e.id === id)[0];
    if (!hint)
        throw `Hint ${id} je požadován před jeho přidáním do scény`;

    return hint;
}

export function getWhiteboardById(scene: IScene, id: string): IHint {
    const wh = scene.whiteboards.filter(e => e.id === id)[0];
    if (!wh)
        throw `Whiteboard ${id} je požadován před jeho přidáním do scény`;

    return wh;
}

export function getVertexById(scene: IScene, id: string): IVertex {
    const vertex = scene.vertices.filter(e => e.id === id)[0];
    if (!vertex)
        throw `Vertex ${id} je požadován před jeho přidáním do scény`;

    return vertex;
}


export function getItemById<TItem extends IItem>(array: TItem[], id: string): TItem {
    const item = array.filter(e => e.id === id)[0];
    if (!item)
        throw `Item ${id} je požadován před jeho přidáním do scény`;

    return item;
}