import * as b from 'bobril';
import * as bs from 'bobrilstrap';
import { linearInterpolation } from '../common/computer';
import * as constants from '../common/constants';
import { svg } from '../components/visualisation/svg';
import { svgWrap } from '../components/visualisation/svgWrap';
import { edge } from '../components/graph/edge';
import { vertexWithLabel } from '../components/graph/vertexWithLabel';
import { proofControls } from '../components/controls/proofControls';
import { buttonLeft } from '../components/controls/buttonLeft';
import { buttonRight } from '../components/controls/buttonRight';
import { buttonStart } from '../components/controls/buttonStart';
import { buttonErase } from '../components/controls/buttonErase';
import { text } from '../components/negations/text';
import { buttonFullScreen } from '../components/controls/buttonFullScreen';
import { buttonNumber } from '../components/controls/buttonNumber';
import { hint } from '../components/visualisation/hint';
import { whiteboard } from '../components/visualisation/whiteboard';
import { axis } from '../components/negations/axis';
import { image } from '../components/negations/image';
import { halfCircle } from '../components/negations/halfCircle';
import { IScene } from '../model/iScene';
import { IItem } from '../model/iItem';
import { IExamples } from '../model/iExamples';
import { IScenario } from '../model/iScenario';
import { cloud } from '../components/graph/cloud';
import * as css from '../theme/css/css';

export interface INegationsViewerData {
    examples: () => IExamples;
    negations?: boolean;
}

export interface INegationsViewerCtx extends b.IBobrilCtx {
    data: INegationsViewerData;
    exampleIndex: number;
    stepIndex: number;
    oldStepIndex: number;
    scenes: IScene[];
    fullScreen: boolean;
    examples: IExamples;
    journey: string;
    oldFullScreen: boolean;
    buttonsEnabled: boolean;
}

export const negationsViewer = b.createVirtualComponent<INegationsViewerData>({
    id: 'negations-viewer',
    init(ctx: INegationsViewerCtx) {
        ctx.buttonsEnabled = true;
        ctx.fullScreen = false;
        createScenes(ctx);
        changeExample(ctx, 0);
        performNextStep(ctx);
    },
    render(ctx: INegationsViewerCtx, me: b.IBobrilNode) {
        me.children = [
            svgWrap({ fullScreen: ctx.fullScreen }, [
                b.style
                    (bs.row({}, [
                        svg({
                            fullScreen: ctx.fullScreen,
                            negations: ctx.data.negations,
                            onJourneyChanged: (journey) => {
                                ctx.journey = journey;
                                b.invalidate(ctx);
                            },
                            journey: ctx.journey
                        },
                            [
                                ctx.scenes[ctx.exampleIndex].cloud ? cloud() : undefined,
                                ctx.scenes[ctx.exampleIndex].halfCircles
                                    .sort(compare)
                                    .map(halfCircleData => b.withKey(
                                        halfCircle({
                                            x: halfCircleData.point.x,
                                            y: halfCircleData.point.y,
                                            rotation: halfCircleData.rotation,
                                            color: halfCircleData.color,
                                            strokeColor: halfCircleData.strokeColor,
                                            strokeWidth: halfCircleData.strokeWidth
                                        }),
                                        'halfCircle' + halfCircleData.id)
                                    ),
                                ctx.scenes[ctx.exampleIndex].vertices
                                    .map(vertexData => b.withKey(
                                        vertexWithLabel({
                                            cx: vertexData.point.x,
                                            cy: vertexData.point.y,
                                            r: 17,
                                            onMove: (dx, dy) => {
                                                if (vertexData.move === undefined) {
                                                    vertexData.point.x += dx;
                                                    vertexData.point.y += dy;
                                                    b.invalidate(ctx);
                                                }
                                            },
                                            strokeColor: vertexData.strokeColor,
                                            strokeWidth: vertexData.strokeWidth,
                                            color: vertexData.color,
                                            label: vertexData.label
                                        }),
                                        'vertex' + vertexData.id)
                                    ),
                                ctx.scenes[ctx.exampleIndex].edges
                                    .sort(compare)
                                    .map(edgeData => b.withKey(
                                        edge({
                                            x1: edgeData.startPoint.x,
                                            y1: edgeData.startPoint.y,
                                            x2: interpolate(edgeData.startPoint.x, edgeData.endPoint.x, edgeData.interpolation),
                                            y2: interpolate(edgeData.startPoint.y, edgeData.endPoint.y, edgeData.interpolation),
                                            strokeWidth: edgeData.strokeWidth,
                                            color: edgeData.color,
                                            dotted: edgeData.dotted
                                        }),
                                        'edge' + edgeData.id)
                                    ),
                                ctx.scenes[ctx.exampleIndex].images
                                    .sort(compare)
                                    .map(imageData => b.withKey(
                                        image({
                                            x: imageData.point.x,
                                            y: imageData.point.y,
                                            width: imageData.width,
                                            height: imageData.height,
                                            type: imageData.type
                                        }),
                                        'image' + imageData.id)
                                    ),
                                ctx.scenes[ctx.exampleIndex].texts
                                    .sort(compare)
                                    .map(textData => b.withKey(
                                        text({
                                            x: textData.point.x,
                                            y: textData.point.y,
                                            backgroundColor: textData.backgroundColor,
                                            color: textData.color,
                                            size: textData.size,
                                            children: textData.text
                                        }),
                                        'text' + textData.id)
                                    ),
                                ctx.scenes[ctx.exampleIndex].axis === undefined ? undefined : axis()
                            ]),

                        ctx.scenes[ctx.exampleIndex].hints
                            .sort(compare)
                            .map((hintData, hintIndex) => b.withKey(
                                hint({
                                    top: hintData.point.y,
                                    left: hintData.point.x,
                                    countedTop: hintData.countedPoint === undefined ? hintData.point.y : hintData.countedPoint.y,
                                    countedLeft: hintData.countedPoint === undefined ? hintData.point.x : hintData.countedPoint.x,
                                    onMove: (dx, dy) => {
                                        hintData.point.x += dx;
                                        hintData.point.y += dy;
                                        b.invalidate(ctx);
                                    },
                                    color: hintData.color,
                                    backgroundColor: hintData.backgroundColor,
                                    fullScreen: ctx.fullScreen,
                                    waitForAnimation: hintData.waitForAnimation,
                                    children: hintData.text
                                }),
                                'hint' + hintData.id)
                            ),
                        ctx.scenes[ctx.exampleIndex].whiteboards
                            .sort(compare)
                            .map((whiteboardData, whiteboardIndex) => b.withKey(
                                whiteboard({
                                    top: whiteboardData.point.y,
                                    left: whiteboardData.point.x,
                                    countedTop: whiteboardData.countedPoint === undefined ?
                                        whiteboardData.point.y : whiteboardData.countedPoint.y,
                                    countedLeft: whiteboardData.countedPoint === undefined ?
                                        whiteboardData.point.x : whiteboardData.countedPoint.x,
                                    color: whiteboardData.color,
                                    margin: whiteboardData.margin,
                                    waitForAnimation: whiteboardData.waitForAnimation,
                                    backgroundColor: whiteboardData.backgroundColor,
                                    fullScreen: ctx.fullScreen,
                                    fontSize: whiteboardData.fontSize,
                                    children: whiteboardData.text
                                }),
                                'whiteboard' + whiteboardData.id)
                            ),
                    ]),
                    ctx.fullScreen ? css.contentRowFullStyle : css.contentRowStyle),
                b.style
                    (bs.row({}, [
                        proofControls({ bottom: 0, left: 0 }, [
                            bs.buttonGroup({}, [
                                buttonLeft({
                                    onClick: () => {
                                        if (ctx.buttonsEnabled) {
                                            rollbackStep(ctx);
                                        }
                                    },
                                    fullScreen: ctx.fullScreen,
                                    enabled: ctx.buttonsEnabled ?
                                        ctx.stepIndex === 0 ? false : true : false
                                }),
                                buttonRight({
                                    onClick: () => {
                                        if (ctx.buttonsEnabled) {
                                            performNextStep(ctx);
                                        }
                                    },
                                    fullScreen: ctx.fullScreen,
                                    enabled: ctx.buttonsEnabled ?
                                        ctx.stepIndex === ctx.examples[ctx.exampleIndex].steps.length - 1 ? false : true : false
                                }),
                                buttonStart({
                                    onClick: () => {
                                        if (ctx.buttonsEnabled) {
                                            resetScene(ctx);
                                        }
                                    },
                                    fullScreen: ctx.fullScreen,
                                    enabled: ctx.buttonsEnabled
                                }),
                                buttonErase({ onClick: () => eraseScene(ctx), fullScreen: ctx.fullScreen }),
                            ]),
                            ctx.examples.length > 1 ? bs.buttonGroup(
                                {},
                                ctx.examples.map((example, index) => buttonNumber({
                                    graphNumber: index,
                                    onClick: () => {
                                        resetScene(ctx);
                                        changeExample(ctx, index);
                                    },
                                    fullScreen: ctx.fullScreen
                                }))
                            ) : undefined,
                            bs.buttonGroup({}, [buttonFullScreen({ onClick: () => toggleFullScreen(ctx), fullScreen: ctx.fullScreen })])
                        ]),

                    ]),
                    css.constrolsRowStyle),
            ]),
        ];
    },
    postUpdateDom(ctx: INegationsViewerCtx) {
        let svg = document.getElementsByTagName('svg')[0];

        ctx.scenes[ctx.exampleIndex].hints.forEach(hintData => {
            if (hintData.countedPoint === undefined) {
                hintData.countedPoint = { x: hintData.point.x, y: hintData.point.y };
                hintData.oldCountedPoint = { x: hintData.point.x, y: hintData.point.y };
            }
            hintData.countedPoint.y = ((hintData.point.y * svg.getBoundingClientRect().height) / constants.height);
            hintData.countedPoint.x = ((hintData.point.x * svg.getBoundingClientRect().width) / constants.width);

            if (hintData.oldCountedPoint.x !== hintData.countedPoint.x ||
                hintData.oldCountedPoint.y !== hintData.countedPoint.y) {
                b.invalidate(ctx);
            }

            hintData.oldCountedPoint.x = hintData.countedPoint.x;
            hintData.oldCountedPoint.y = hintData.countedPoint.y;
        });

        ctx.scenes[ctx.exampleIndex].whiteboards.forEach(whiteboardData => {
            if (whiteboardData.countedPoint === undefined) {
                whiteboardData.countedPoint = { x: whiteboardData.point.x, y: whiteboardData.point.y };
                whiteboardData.oldCountedPoint = { x: whiteboardData.point.x, y: whiteboardData.point.y };
            }
            whiteboardData.countedPoint.y = ((whiteboardData.point.y * svg.getBoundingClientRect().height) / constants.height);
            whiteboardData.countedPoint.x = ((whiteboardData.point.x * svg.getBoundingClientRect().width) / constants.width);

            if (whiteboardData.oldCountedPoint.x !== whiteboardData.countedPoint.x ||
                whiteboardData.oldCountedPoint.y !== whiteboardData.countedPoint.y) {
                b.invalidate(ctx);
            }

            whiteboardData.oldCountedPoint.x = whiteboardData.countedPoint.x;
            whiteboardData.oldCountedPoint.y = whiteboardData.countedPoint.y;
        });
    }
});

function toggleFullScreen(ctx: INegationsViewerCtx) {
    ctx.fullScreen = !ctx.fullScreen; b.invalidate();
}

function enableButtons(ctx: INegationsViewerCtx) {
    ctx.buttonsEnabled = true;
    b.invalidate(ctx);
}

function changeExample(ctx: INegationsViewerCtx, index: number) {

    ctx.exampleIndex = index;
    b.invalidate(ctx);
}

function createScenes(ctx: INegationsViewerCtx) {
    ctx.examples = ctx.data.examples();
    ctx.stepIndex = -1;
    ctx.scenes = [];
    ctx.examples.forEach(example => ctx.scenes.push({
        edges: [], hints: [], vertices: [],
        whiteboards: [], axis: undefined, halfCircles: [], images: [], texts: [], cloud: undefined
    }));
}

function resetScene(ctx: INegationsViewerCtx, invalidate: boolean = false) {
    while (ctx.stepIndex !== 0) {
        rollbackStep(ctx);
    }
    ctx.examples[ctx.exampleIndex] = ctx.data.examples()[ctx.exampleIndex];
    ctx.scenes[ctx.exampleIndex] = {
        edges: [], hints: [], vertices: [],
        whiteboards: [], axis: undefined, halfCircles: [], images: [], texts: [], cloud: undefined
    };
    performActions(ctx.examples[ctx.exampleIndex], ctx.stepIndex, ctx.scenes[ctx.exampleIndex]);
    b.invalidate(ctx);
}

function eraseScene(ctx: INegationsViewerCtx, invalidate: boolean = false) {
    ctx.journey = '';
    b.invalidate(ctx);
}

function performNextStep(ctx: INegationsViewerCtx) {
    if (ctx.stepIndex >= ctx.examples[ctx.exampleIndex].steps.length - 1)
        return;

    ctx.stepIndex++;
    ctx.scenes.forEach((scene, index) => performActions(ctx.examples[index], ctx.stepIndex, scene));
    b.invalidate(ctx);
}

function performActions(scenario: IScenario, stepIndex: number, scene: IScene) {
    scenario.steps[stepIndex].actions.forEach(action => action.update(scene));
}

function rollbackActions(scenario: IScenario, stepIndex: number, scene: IScene) {
    [...scenario.steps[stepIndex].actions].reverse().forEach(action => action.rollback(scene));
}

function rollbackStep(ctx: INegationsViewerCtx) {
    if (ctx.stepIndex <= 0)
        return;

    ctx.scenes.forEach((scene, index) => rollbackActions(ctx.examples[index], ctx.stepIndex, scene));
    ctx.stepIndex--;
    b.invalidate(ctx);
}

function interpolate(a: number, b: number, ratio: number) {
    if (ratio) {
        return linearInterpolation(a, b, ratio);
    } else {
        return b;
    }
}

function compare<TItem extends IItem>(a: TItem, b: TItem) {
    const priorityA = a.priority || 0;
    const priorityB = b.priority || 0;
    if (priorityA > priorityB) {
        return 1;
    } else if (priorityA < priorityB) {
        return -1;
    }
    return 0;
}