import * as b from 'bobril';
import * as bs from 'bobrilstrap';
import { linearInterpolation } from '../common/computer';
import * as constants from '../common/constants';
import { svg } from '../components/visualisation/svg';
import { svgWrap } from '../components/visualisation/svgWrap';
import { edge } from '../components/graph/edge';
import { vertexWithLabel } from '../components/graph/vertexWithLabel';
import { proofControls } from '../components/controls/proofControls';
import { slider } from '../components/controls/slider';
import { buttonLeft } from '../components/controls/buttonLeft';
import { progressBar } from '../components/controls/progressBar';
import { buttonRight } from '../components/controls/buttonRight';
import { buttonStart } from '../components/controls/buttonStart';
import { buttonErase } from '../components/controls/buttonErase';
import { buttonRepeat } from '../components/controls/buttonRepeat';
import { text } from '../components/negations/text';
import { buttonFullScreen } from '../components/controls/buttonFullScreen';
import { buttonNumber } from '../components/controls/buttonNumber';
import { hint } from '../components/visualisation/hint';
import { whiteboard } from '../components/visualisation/whiteboard';
import { definitions } from '../components/proofs/definitions';
import { axis } from '../components/negations/axis';
import { image } from '../components/negations/image';
import { halfCircle } from '../components/negations/halfCircle';
import { IScene } from '../model/iScene';
import { IItem } from '../model/iItem';
import { IExamples } from '../model/iExamples';
import { IScenario } from '../model/iScenario';
import { sentence } from '../components/proofs/sentence';
import { cloud } from '../components/graph/cloud';
import { buttonSentences } from '../components/controls/buttonSentences';
import * as css from '../theme/css/css';

export interface IViewerData {
    examples: () => IExamples;
    cloud?: boolean;
    negations?: boolean;
}

export interface IViewerCtx extends b.IBobrilCtx {
    data: IViewerData;
    exampleIndex: number;
    stepIndex: number;
    oldStepIndex: number;
    scenes: IScene[];
    fullScreen: boolean;
    examples: IExamples;
    journey: string;
    num: number;
    oldHeight: number;
    oldWidth: number;
    oldFullScreen: boolean;
    shownSentences: boolean;
    buttonsEnabled: boolean;
    repeated: boolean;
    repeatedOnWayBack: boolean;
    animationRunning: boolean;
}

export const viewer = b.createVirtualComponent<IViewerData>({
    id: 'viewer',
    init(ctx: IViewerCtx) {
        ctx.buttonsEnabled = true;
        ctx.fullScreen = false;
        createScenes(ctx);
        changeExample(ctx, 0);
        performNextStep(ctx);
    },
    render(ctx: IViewerCtx, me: b.IBobrilNode) {
        me.children = [
            svgWrap({ fullScreen: ctx.fullScreen }, [
                b.style
                    (bs.row({}, [
                        b.style(bs.col({ size: bs.Size.Xs, span: 8 }, [
                            buttonSentences({
                                onClick: () => showSentences(ctx),
                                fullScreen: ctx.fullScreen, shownSentences: ctx.shownSentences
                            }),
                            svg({
                                fullScreen: ctx.fullScreen,
                                negations: ctx.data.negations,
                                onJourneyChanged: (journey) => {
                                    ctx.journey = journey;
                                    b.invalidate(ctx);
                                },
                                journey: ctx.journey,
                                shownSentences: ctx.shownSentences
                            },
                                [
                                    ctx.scenes[ctx.exampleIndex].cloud ? cloud() : undefined,
                                    ctx.scenes[ctx.exampleIndex].halfCircles
                                        .sort(compare)
                                        .map(halfCircleData => b.withKey(
                                            halfCircle({
                                                x: halfCircleData.point.x,
                                                y: halfCircleData.point.y,
                                                rotation: halfCircleData.rotation,
                                                color: halfCircleData.color,
                                                strokeColor: halfCircleData.strokeColor,
                                                strokeWidth: halfCircleData.strokeWidth
                                            }),
                                            'halfCircle' + halfCircleData.id)
                                        ),
                                    ctx.scenes[ctx.exampleIndex].edges
                                        .sort(compare)
                                        .map(edgeData => b.withKey(
                                            edge({
                                                x1: edgeData.startPoint.x,
                                                y1: edgeData.startPoint.y,
                                                x2: interpolate(edgeData.startPoint.x, edgeData.endPoint.x, edgeData.interpolation),
                                                y2: interpolate(edgeData.startPoint.y, edgeData.endPoint.y, edgeData.interpolation),
                                                strokeWidth: edgeData.strokeWidth,
                                                color: edgeData.color,
                                                dotted: edgeData.dotted,
                                                moved: edgeData.moved,
                                                waitForAnimation: edgeData.waitForAnimation,
                                                stepIndex: ctx.stepIndex,
                                                oldStepIndex: ctx.oldStepIndex
                                            }),
                                            'edge' + edgeData.id)
                                        ),
                                    ctx.scenes[ctx.exampleIndex].vertices
                                        .map(vertexData => b.withKey(
                                            vertexWithLabel({
                                                cx: vertexData.point.x,
                                                cy: vertexData.point.y,
                                                r: 17,
                                                onMove: (dx, dy) => {
                                                    if (vertexData.move === undefined) {
                                                        vertexData.point.x += dx;
                                                        vertexData.point.y += dy;
                                                        b.invalidate(ctx);
                                                    }
                                                },
                                                strokeColor: vertexData.strokeColor,
                                                strokeWidth: vertexData.strokeWidth,
                                                color: vertexData.color,
                                                label: vertexData.label
                                            }),
                                            'vertex' + vertexData.id)
                                        ),
                                    ctx.scenes[ctx.exampleIndex].images
                                        .sort(compare)
                                        .map(imageData => b.withKey(
                                            image({
                                                x: imageData.point.x,
                                                y: imageData.point.y,
                                                width: imageData.width,
                                                height: imageData.height,
                                                type: imageData.type
                                            }),
                                            'image' + imageData.id)
                                        ),
                                    ctx.scenes[ctx.exampleIndex].texts
                                        .sort(compare)
                                        .map(textData => b.withKey(
                                            text({
                                                x: textData.point.x,
                                                y: textData.point.y,
                                                backgroundColor: textData.backgroundColor,
                                                color: textData.color,
                                                size: textData.size,
                                                children: textData.text
                                            }),
                                            'text' + textData.id)
                                        ),
                                    ctx.scenes[ctx.exampleIndex].axis === undefined ? undefined : axis()
                                ]),
                            ctx.scenes[ctx.exampleIndex].hints
                                .sort(compare)
                                .map((hintData, hintIndex) => b.withKey(
                                    hint({
                                        top: hintData.point.y,
                                        left: hintData.point.x,
                                        countedTop: hintData.countedPoint === undefined ? hintData.point.y : hintData.countedPoint.y,
                                        countedLeft: hintData.countedPoint === undefined ? hintData.point.x : hintData.countedPoint.x,
                                        onMove: (dx, dy) => {
                                            hintData.point.x += dx;
                                            hintData.point.y += dy;
                                            b.invalidate(ctx);
                                        },
                                        color: hintData.color,
                                        backgroundColor: hintData.backgroundColor,
                                        fullScreen: ctx.fullScreen,
                                        waitForAnimation: hintData.waitForAnimation,
                                        children: hintData.text
                                    }),
                                    'hint' + hintData.id)
                                ),
                            ctx.scenes[ctx.exampleIndex].whiteboards
                                .sort(compare)
                                .map((whiteboardData, whiteboardIndex) => b.withKey(
                                    whiteboard({
                                        top: whiteboardData.point.y,
                                        left: whiteboardData.point.x,
                                        countedTop: whiteboardData.countedPoint === undefined ?
                                            whiteboardData.point.y : whiteboardData.countedPoint.y,
                                        countedLeft: whiteboardData.countedPoint === undefined ?
                                            whiteboardData.point.x : whiteboardData.countedPoint.x,
                                        color: whiteboardData.color,
                                        margin: whiteboardData.margin,
                                        waitForAnimation: whiteboardData.waitForAnimation,
                                        backgroundColor: whiteboardData.backgroundColor,
                                        fullScreen: ctx.fullScreen,
                                        fontSize: whiteboardData.fontSize,
                                        children: whiteboardData.text
                                    }),
                                    'whiteboard-' + whiteboardData.id)
                                ),
                            sentence({
                                color: !!ctx.examples[ctx.exampleIndex].steps[ctx.stepIndex].sentence
                                && ctx.examples[ctx.exampleIndex].steps[ctx.stepIndex].sentence.color,
                                text: !!ctx.examples[ctx.exampleIndex].steps[ctx.stepIndex].sentence
                                && ctx.examples[ctx.exampleIndex].steps[ctx.stepIndex].sentence.text,
                                shownSentences: ctx.shownSentences,
                                fullScreen: ctx.fullScreen
                            }),
                        ]), { height: '100%', paddingLeft: 0, cssFloat: 'right' }),
                        ctx.examples[ctx.exampleIndex].definitions !== undefined ?
                            b.style(bs.col({ size: bs.Size.Xs, span: 4 }, [
                                definitions({
                                    step: ctx.stepIndex, defs: ctx.examples[ctx.exampleIndex].definitions
                                }, [])]
                            ), { height: '100%', paddingLeft: 0 }) : undefined,
                    ]),
                    ctx.fullScreen ? css.contentRowFullStyle : css.contentRowStyle),
                b.style
                    (bs.row({}, [
                        proofControls({ bottom: 5, left: 0 }, [
                            bs.buttonGroup({}, [
                                buttonRepeat({
                                    onClick: () => {
                                        if (ctx.buttonsEnabled) {
                                            repeatScene(ctx);
                                        }
                                    },
                                    fullScreen: ctx.fullScreen,
                                    enabled: ctx.buttonsEnabled
                                }),
                                buttonLeft({
                                    onClick: () => {
                                        if (ctx.buttonsEnabled) {
                                            rollbackStep(ctx);
                                        }
                                    },
                                    fullScreen: ctx.fullScreen,
                                    enabled: ctx.buttonsEnabled ?
                                        ctx.stepIndex === 0 ? false : true : false
                                }),
                            ]),
                            progressBar({
                                onePercentValue:
                                ctx.examples[ctx.exampleIndex].steps[ctx.stepIndex].time !== undefined
                                    ? ctx.oldStepIndex > ctx.stepIndex
                                        ? undefined
                                        : ctx.examples[ctx.exampleIndex].steps[ctx.stepIndex].time / 100
                                    : undefined,
                                onDone: () => {
                                    enableButtons(ctx);
                                    ctx.animationRunning = false;
                                },
                                onStarted: () => {
                                    ctx.animationRunning = true;
                                }
                            }),
                            bs.buttonGroup({}, [
                                buttonRight({
                                    onClick: () => {
                                        if (ctx.buttonsEnabled) {
                                            performNextStep(ctx);
                                        }
                                    },
                                    fullScreen: ctx.fullScreen,
                                    enabled: ctx.buttonsEnabled ?
                                        ctx.stepIndex === ctx.examples[ctx.exampleIndex].steps.length - 1 ? false : true : false
                                }),
                                buttonStart({
                                    onClick: () => {
                                        if (ctx.buttonsEnabled) {
                                            resetScene(ctx);
                                        }
                                    },
                                    fullScreen: ctx.fullScreen,
                                    enabled: ctx.buttonsEnabled
                                }),
                                buttonErase({ onClick: () => eraseScene(ctx), fullScreen: ctx.fullScreen }),
                                buttonFullScreen({ onClick: () => toggleFullScreen(ctx), fullScreen: ctx.fullScreen }),
                            ]),
                            slider({
                                id: 'slider1', value: constants.slowedTime, onChanged: (value) => {
                                    constants.setSlowedTime(value);
                                    b.invalidate(ctx);
                                }, fullScreen: ctx.fullScreen, shownSentences: ctx.shownSentences, enabled: ctx.buttonsEnabled
                            }),
                            ctx.examples.length > 1 ? bs.buttonGroup(
                                {},
                                ctx.examples.map((example, index) => buttonNumber({
                                    graphNumber: index,
                                    onClick: () => changeExample(ctx, index),
                                    fullScreen: ctx.fullScreen
                                }))
                            ) : undefined])
                    ]),
                    css.constrolsRowStyle),
            ])
        ];
    },
    postUpdateDom(ctx: IViewerCtx) {
        let svg = document.getElementsByTagName('svg')[0];

        ctx.scenes[ctx.exampleIndex].hints.forEach(hintData => {
            if (hintData.countedPoint === undefined) {
                hintData.countedPoint = { x: hintData.point.x, y: hintData.point.y };
                hintData.oldCountedPoint = { x: hintData.point.x, y: hintData.point.y };
            }
            hintData.countedPoint.y = (
                (hintData.point.y * svg.getBoundingClientRect().height)
                / constants.height
            );
            hintData.countedPoint.x = (
                (hintData.point.x * svg.getBoundingClientRect().width)
                / constants.width
            );

            if (hintData.oldCountedPoint.x !== hintData.countedPoint.x ||
                hintData.oldCountedPoint.y !== hintData.countedPoint.y) {
                b.invalidate(ctx);
            }

            hintData.oldCountedPoint.x = hintData.countedPoint.x;
            hintData.oldCountedPoint.y = hintData.countedPoint.y;
        });

        ctx.scenes[ctx.exampleIndex].whiteboards.forEach(whiteboardData => {
            if (whiteboardData.countedPoint === undefined) {
                whiteboardData.countedPoint = { x: whiteboardData.point.x, y: whiteboardData.point.y };
                whiteboardData.oldCountedPoint = { x: whiteboardData.point.x, y: whiteboardData.point.y };
            }
            whiteboardData.countedPoint.y = ((whiteboardData.point.y * svg.getBoundingClientRect().height) / constants.height);
            whiteboardData.countedPoint.x = ((whiteboardData.point.x * svg.getBoundingClientRect().width) / constants.width);

            if (whiteboardData.oldCountedPoint.x !== whiteboardData.countedPoint.x ||
                whiteboardData.oldCountedPoint.y !== whiteboardData.countedPoint.y) {
                b.invalidate(ctx);
            }

            whiteboardData.oldCountedPoint.x = whiteboardData.countedPoint.x;
            whiteboardData.oldCountedPoint.y = whiteboardData.countedPoint.y;
        });
        if (ctx.repeated) {
            ctx.scenes.forEach((scene, index) => performActions(ctx.examples[index], ctx.stepIndex, scene));

            if (!ctx.repeatedOnWayBack) {
                ctx.scenes.forEach((scene, index) => ctx.examples[index].steps[ctx.stepIndex].time
                    = ctx.examples[index].steps[ctx.stepIndex].time + 5); 
            }

            ctx.repeated = false;
            b.invalidate(ctx);
        }
    }
});

function toggleFullScreen(ctx: IViewerCtx) {
    ctx.fullScreen = !ctx.fullScreen;
    b.invalidate();
}

function showSentences(ctx: IViewerCtx) {
    ctx.shownSentences = !ctx.shownSentences;
    b.invalidate(ctx);
}

function enableButtons(ctx: IViewerCtx) {
    ctx.buttonsEnabled = true;
    b.invalidate(ctx);
}

function changeExample(ctx: IViewerCtx, index: number) {
    ctx.exampleIndex = index;
    b.invalidate(ctx);
}

function createScenes(ctx: IViewerCtx) {
    ctx.examples = ctx.data.examples();
    ctx.stepIndex = -1;
    ctx.scenes = [];
    ctx.examples.forEach(example => ctx.scenes.push({
        edges: [], hints: [], vertices: [],
        whiteboards: [], axis: undefined, halfCircles: [], images: [], texts: [], cloud: undefined
    }));
}

function resetScene(ctx: IViewerCtx, invalidate: boolean = false) {
    while (ctx.stepIndex !== 0) {
        rollbackStep(ctx);
    }
    ctx.examples[ctx.exampleIndex] = ctx.data.examples()[ctx.exampleIndex];
    ctx.scenes[ctx.exampleIndex] = {
        edges: [], hints: [], vertices: [],
        whiteboards: [], axis: undefined, halfCircles: [], images: [], texts: [], cloud: undefined
    };

    if (ctx.animationRunning) {
        let event = document.createEvent('Event');
        event.initEvent('finish-animation', true, true);
        document.body.dispatchEvent(event);

        ctx.animationRunning = false;
    }

    performActions(ctx.examples[ctx.exampleIndex], ctx.stepIndex, ctx.scenes[ctx.exampleIndex]);
    b.invalidate(ctx);
}

function repeatScene(ctx: IViewerCtx, invalidate: boolean = false) {
    if (ctx.stepIndex < 0)
        return;

    if (ctx.oldStepIndex > ctx.stepIndex) {
        ctx.repeatedOnWayBack = true;
    } else {
        ctx.repeatedOnWayBack = false;
    }
    ctx.oldStepIndex = ctx.stepIndex;

    if (ctx.animationRunning) {
        let event = document.createEvent('Event');
        event.initEvent('finish-animation', true, true);
        document.body.dispatchEvent(event);

        ctx.animationRunning = false;
    }

    ctx.scenes.forEach((scene, index) => rollbackActions(ctx.examples[index], ctx.stepIndex, scene));
    ctx.repeated = true;
    b.invalidate(ctx);
}

function eraseScene(ctx: IViewerCtx, invalidate: boolean = false) {
    ctx.journey = '';
    b.invalidate(ctx);
}

function performNextStep(ctx: IViewerCtx) {
    if (ctx.stepIndex >= ctx.examples[ctx.exampleIndex].steps.length - 1)
        return;

    if (ctx.animationRunning) {
        let event = document.createEvent('Event');
        event.initEvent('finish-animation', true, true);
        document.body.dispatchEvent(event);

        ctx.animationRunning = false;
    }

    ctx.oldStepIndex = ctx.stepIndex;
    ctx.stepIndex++;
    ctx.scenes.forEach((scene, index) => performActions(ctx.examples[index], ctx.stepIndex, scene));
    b.invalidate(ctx);
}

function performActions(scenario: IScenario, stepIndex: number, scene: IScene) {
    scenario.steps[stepIndex].actions.forEach(action => action.update(scene));
}

function rollbackActions(scenario: IScenario, stepIndex: number, scene: IScene) {
    [...scenario.steps[stepIndex].actions].reverse().forEach(action => action.rollback(scene));
}

function rollbackStep(ctx: IViewerCtx) {
    if (ctx.stepIndex <= 0)
        return;

    if (ctx.animationRunning) {
        let event = document.createEvent('Event');
        event.initEvent('finish-animation', true, true);
        document.body.dispatchEvent(event);

        ctx.animationRunning = false;
    }

    ctx.scenes.forEach((scene, index) => rollbackActions(ctx.examples[index], ctx.stepIndex, scene));
    ctx.oldStepIndex = ctx.stepIndex;
    ctx.stepIndex--;
    b.invalidate(ctx);
}

function interpolate(a: number, b: number, ratio: number) {
    if (ratio) {
        return linearInterpolation(a, b, ratio);
    } else {
        return b;
    }
}

function compare<TItem extends IItem>(a: TItem, b: TItem) {
    const priorityA = a.priority || 0;
    const priorityB = b.priority || 0;
    if (priorityA > priorityB) {
        return 1;
    } else if (priorityA < priorityB) {
        return -1;
    }
    return 0;
}